# w.c.s. - web application for online forms
# Copyright (C) 2005-2020  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from quixote import get_publisher

from wcs.qommon import get_logger
from wcs.qommon.xml_storage import XmlStorableObject


class MailTemplate(XmlStorableObject):
    _names = 'mail-templates'
    xml_root_node = 'mail-template'
    backoffice_class = 'wcs.admin.mail_templates.MailTemplatePage'

    name = None
    slug = None
    description = None
    subject = None
    body = None
    attachments = []

    # declarations for serialization
    XML_NODES = [
        ('name', 'str'),
        ('slug', 'str'),
        ('description', 'str'),
        ('subject', 'str'),
        ('body', 'str'),
        ('attachments', 'str_list'),
    ]

    def __init__(self, name=None):
        XmlStorableObject.__init__(self)
        self.name = name

    def get_admin_url(self):
        base_url = get_publisher().get_backoffice_url()
        return '%s/workflows/mail-templates/%s/' % (base_url, self.id)

    def store(self, comment=None, *args, **kwargs):
        assert not self.is_readonly()
        if self.slug is None:
            # set slug if it's not yet there
            self.slug = self.get_new_slug()
        super().store(*args, **kwargs)
        if get_publisher().snapshot_class:
            get_publisher().snapshot_class.snap(instance=self, comment=comment)

    def get_places_of_use(self):
        from wcs.workflows import Workflow

        for workflow in Workflow.select(ignore_errors=True, ignore_migration=True):
            for item in workflow.get_all_items():
                if item.key != 'sendmail':
                    continue
                if item.mail_template == self.slug:
                    yield workflow
                    break

    def is_in_use(self):
        return any(self.get_places_of_use())

    @classmethod
    def get_as_options_list(cls):
        options = []
        for mail_template in cls.select(order_by='name'):
            options.append((mail_template.slug, mail_template.name, mail_template.slug))
        return options

    @classmethod
    def get_by_slug(cls, slug):
        mail_template = super().get_by_slug(slug, ignore_errors=True)
        if mail_template is None:
            get_logger().warning("mail template '%s' does not exist" % slug)
        return mail_template
