# w.c.s. - web application for online forms
# Copyright (C) 2005-2021  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import collections

from django.http import HttpResponseBadRequest, HttpResponseForbidden, JsonResponse
from django.urls import reverse
from django.views.generic import View
from quixote import get_publisher

from wcs import sql
from wcs.api_utils import is_url_signed
from wcs.backoffice.management import FormPage
from wcs.categories import Category
from wcs.formdef import FormDef
from wcs.qommon import _, misc
from wcs.qommon.misc import C_
from wcs.qommon.storage import Equal, NotEqual, Or


class RestrictedView(View):
    def dispatch(self, *args, **kwargs):
        if not is_url_signed():
            return HttpResponseForbidden()
        return super().dispatch(*args, **kwargs)


class IndexView(RestrictedView):
    def get(self, request, *args, **kwargs):
        if not get_publisher().is_using_postgresql():
            return JsonResponse({'data': [], 'err': 0})
        categories = Category.select()
        categories.sort(key=lambda x: misc.simplify(x.name))
        category_options = [{'id': '_all', 'label': C_('categories|All')}] + [
            {'id': x.id, 'label': x.name} for x in categories
        ]
        forms = FormDef.select()
        forms.sort(key=lambda x: misc.simplify(x.name))
        form_options = [{'id': '_all', 'label': _('All')}] + [
            {'id': x.url_name, 'label': x.name} for x in forms
        ]
        return JsonResponse(
            {
                'data': [
                    {
                        'name': _('Forms Count'),
                        'url': request.build_absolute_uri(reverse('api-statistics-forms-count')),
                        'id': 'forms_counts',
                        'filters': [
                            {
                                'id': 'time_interval',
                                'label': _('Interval'),
                                'options': [
                                    {
                                        'id': 'month',
                                        'label': _('Month'),
                                    },
                                    {
                                        'id': 'year',
                                        'label': _('Year'),
                                    },
                                    {
                                        'id': 'weekday',
                                        'label': _('Week day'),
                                    },
                                    {
                                        'id': 'hour',
                                        'label': _('Hour'),
                                    },
                                ],
                                'required': True,
                                'default': 'month',
                            },
                            {
                                'id': 'category',
                                'label': _('Category'),
                                'options': category_options,
                                'required': True,
                                'default': '_all',
                            },
                            {
                                'id': 'form',
                                'label': _('Form'),
                                'options': form_options,
                                'required': True,
                                'default': '_all',
                                'has_subfilters': True,
                            },
                        ],
                    }
                ]
            }
        )


class FormsCountView(RestrictedView):
    def get(self, request, *args, **kwargs):
        time_interval = request.GET.get('time_interval', 'month')
        totals_kwargs = {
            'period_start': request.GET.get('start'),
            'period_end': request.GET.get('end'),
            'criterias': [],
        }
        category_id = request.GET.get('category', '_all')
        formdef_slug = request.GET.get('form', '_all')
        group_by = request.GET.get('group-by')
        subfilters = []
        if formdef_slug != '_all':
            try:
                formdef = FormDef.get_by_urlname(formdef_slug, ignore_migration=True)
            except KeyError:
                return HttpResponseBadRequest('invalid form')
            form_page = FormPage(formdef=formdef, update_breadcrumbs=False)

            totals_kwargs['criterias'].append(Equal('formdef_id', formdef.id))
            totals_kwargs['criterias'].extend(self.get_filters_criterias(formdef, form_page))
            if group_by:
                totals_kwargs['group_by'] = self.get_group_by_column(form_page, group_by)
            subfilters = self.get_subfilters(form_page)
        elif category_id != '_all':
            totals_kwargs['criterias'].append(Equal('category_id', category_id))

        time_interval_methods = {
            'month': sql.get_monthly_totals,
            'year': sql.get_yearly_totals,
            'weekday': sql.get_weekday_totals,
            'hour': sql.get_hour_totals,
        }
        if time_interval in time_interval_methods:
            totals = time_interval_methods[time_interval](**totals_kwargs)
        else:
            return HttpResponseBadRequest('invalid time_interval parameter')

        if not totals_kwargs.get('group_by'):
            x_labels = [x[0] for x in totals]
            series = [{'label': _('Forms Count'), 'data': [x[1] for x in totals]}]
        else:
            x_labels, series = self.get_grouped_data(totals, group_by, formdef)

        return JsonResponse(
            {'data': {'x_labels': x_labels, 'series': series, 'subfilters': subfilters}, 'err': 0}
        )

    def get_filters_criterias(self, formdef, form_page):
        criterias = form_page.get_criterias_from_query()

        selected_status = self.request.GET.get('filter-status')
        applied_filters = None
        if selected_status and selected_status != '_all':
            if selected_status == 'pending':
                applied_filters = ['wf-%s' % x.id for x in formdef.workflow.get_not_endpoint_status()]
            elif selected_status == 'done':
                applied_filters = ['wf-%s' % x.id for x in formdef.workflow.get_endpoint_status()]
            else:
                try:
                    formdef.workflow.get_status(selected_status)
                    applied_filters = ['wf-%s' % selected_status]
                except KeyError:
                    pass

        if applied_filters:
            criterias.append(Or([Equal('status', x) for x in applied_filters]))
        else:
            criterias = [NotEqual('status', 'draft')] + criterias

        return criterias

    @staticmethod
    def get_subfilters(form_page):
        subfilters = []
        field_choices = []
        for field in form_page.get_formdef_fields():
            if not getattr(field, 'include_in_statistics', False) or not field.contextual_varname:
                continue

            field_key = 'filter-%s' % field.contextual_varname
            field.required = False

            if field.type == 'status':
                waitpoint_status = form_page.formdef.workflow.get_waitpoint_status()
                if not waitpoint_status:
                    continue

                field.required = True
                field.default_filter_value = '_all'
                options = [
                    ('_all', _('All')),
                    ('pending', C_('statistics|Open')),
                    ('done', _('Done')),
                ]
                for status in waitpoint_status:
                    options.append((status.id, status.name))
            elif field.type in ('item', 'items'):
                if not get_publisher().is_using_postgresql():
                    continue
                options = form_page.get_item_filter_options(field, selected_filter='all')
                if not options:
                    continue
            elif field.type == 'bool':
                options = [('true', _('Yes')), ('false', _('No'))]
            else:
                continue

            filter_description = {
                'id': field_key,
                'label': field.label,
                'options': [{'id': x[0], 'label': x[1]} for x in options],
                'required': field.required,
            }
            if hasattr(field, 'default_filter_value'):
                filter_description['default'] = field.default_filter_value

            subfilters.append(filter_description)
            if not hasattr(field, 'block_field'):
                field_choices.append((field.contextual_varname, field.label))

        if field_choices:
            subfilters.insert(
                0,
                {
                    'id': 'group-by',
                    'label': _('Group by'),
                    'options': [{'id': x[0], 'label': x[1]} for x in field_choices],
                },
            )

        return subfilters

    def get_group_by_column(self, form_page, group_by):
        if group_by == 'status':
            return group_by
        else:
            fields = [
                x
                for x in form_page.get_formdef_fields()
                if getattr(x, 'contextual_varname', None) == group_by
            ]
            if fields:
                field = fields[0]
                if not hasattr(field, 'block_field'):  # block fields are not supported
                    return sql.get_field_id(field)

    def get_grouped_data(self, totals, group_by, formdef):
        totals_by_time = collections.OrderedDict(
            # time1: {group1: total_11, group2: total_12},
            # time2: {group1: total_21}
        )
        seen_group_values = set(
            # group1, group2
        )
        for total in totals:
            totals_by_group = totals_by_time.setdefault(total[0], collections.Counter())
            if len(total) == 2:
                # ignore empty value used to fill time gaps
                continue
            groups = total[1]
            if not isinstance(groups, list):
                groups = [groups]
            for group in groups:
                totals_by_group[group] += total[2]
                seen_group_values.add(group)

        totals_by_group = {
            # group1: [total_11, total_21],
            # group2: [total_12, None],
        }
        for group in seen_group_values:
            totals_by_group[group] = [totals.get(group) for totals in totals_by_time.values()]

        def get_label(group):
            if group_by == 'status':
                status_id = group.split('-')[1]
                status = formdef.workflow.get_status(status_id)
                return status.name
            group_labels = {True: _('Yes'), False: _('No'), None: _('None')}
            return group_labels.get(group, group)

        x_labels = list(totals_by_time)
        series = [{'label': get_label(group), 'data': data} for group, data in totals_by_group.items()]
        return x_labels, series
